package br.usp.ime.mac5855.knn;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * @author Victor Williams Stafusa da Silva
 */
public final class KnnPonderado<E> {

    private final Map<E, Double> elementosConhecidosComPreco;
    private final KnnStrategy<E> strategy;

    public KnnPonderado(KnnStrategy<E> strategy) {
        this(null, strategy);
    }

    public KnnPonderado(Map<E, Double> elementos, KnnStrategy<E> strategy) {
        if (strategy == null) {
            throw new IllegalArgumentException("O strategy  nulo.");
        }

        this.strategy = strategy;

        this.elementosConhecidosComPreco = new HashMap<E, Double>();
        if (elementos != null) {
            addAll(elementos);
        }
    }

    public void add(E elemento, double preco) {
        elementosConhecidosComPreco.put(elemento, preco);
    }

    public void addAll(Map<E, Double> elementos) {
        for (Map.Entry<E, Double> entry : elementos.entrySet()) {
            this.elementosConhecidosComPreco.put(entry.getKey(), entry.getValue());
        }
    }

    private Map<Double, List<E>> vizinhosPorDistancia(E elemento) {

        Map<Double, List<E>> vizinhos = new TreeMap<Double, List<E>>();

        for (E conhecido : elementosConhecidosComPreco.keySet()) {
            double distancia = strategy.getComparador().distancia(elemento, conhecido);
            List<E> lista = vizinhos.get(distancia);
            if (lista == null) {
                lista = new ArrayList<E>();
            }
            lista.add(conhecido);
            vizinhos.put(distancia, lista);
        }

        return vizinhos;
    }

    public double estimarPreco(E elemento) {

        Map<Double, List<E>> vizinhos = vizinhosPorDistancia(elemento);

        int iteracoesFaltando = strategy.getNumeroVizinhos();
        double estimativa = 0.0;
        double pesoTotal = 0.0;

        a: for (Map.Entry<Double, List<E>> entry : vizinhos.entrySet()) {
            for (E elem : entry.getValue()) {
                double peso = strategy.getPonderador().ponderar(entry.getKey());
                pesoTotal += peso;
                estimativa += elementosConhecidosComPreco.get(elem) * peso;
                iteracoesFaltando--;
                if (iteracoesFaltando == 0) break a;
            }
        }

        return pesoTotal == 0.0 ? 0.0 : estimativa / pesoTotal;
    }

    public double probabilidadePrecoFaixa(E elemento, double precoMinimo, double precoMaximo) {
        Map<Double, List<E>> vizinhos = vizinhosPorDistancia(elemento);

        int iteracoesFaltando = strategy.getNumeroVizinhos();
        double pesoTotal = 0.0;
        double pesoTotalNaFaixa = 0.0;

        a: for (Map.Entry<Double, List<E>> entry : vizinhos.entrySet()) {
            for (E elem : entry.getValue()) {
                double peso = strategy.getPonderador().ponderar(entry.getKey());
                pesoTotal += peso;
                double preco = elementosConhecidosComPreco.get(elem);
                if (precoMinimo <= preco && preco <= precoMaximo) {
                    pesoTotalNaFaixa += peso;
                }
                iteracoesFaltando--;
                if (iteracoesFaltando == 0) break a;
            }
        }

        return pesoTotal == 0.0 ? 0.0 : pesoTotalNaFaixa / pesoTotal;
    }
}
